#include <iostream>
#include <algorithm> // sort()

#include "Reaction.hpp"
#include "fonctions.hpp"
#include "constantes.hpp"


// définir les classe de la std qui seront utilisées
using std::cerr;
using std::endl;

Reaction::Reaction(vector<string> const& infoReaction)
{
    _ec = infoReaction[0];
    _nomEnzyme = infoReaction[1];
    _organisme = infoReaction[2];
    _reversible = infoReaction[5] == "r";
    
    // Stockage des substrats ET leurs stochiometrie
    vector<string> substrats = split(infoReaction[3], "-+-");
    StockageMetabolite(substrats, "substrats");
        
    // Stockage des produits ET leurs stochiometrie
    vector<string> produits = split(infoReaction[4], "-+-");
    StockageMetabolite(produits, "produits");

    remplissageConstante(_km, infoReaction[6], ';');
    remplissageConstante(_kcat, infoReaction[7], ';');
    remplissage_pH(infoReaction[8], '-');
}

Reaction::Reaction(vector<string> const& infoReaction,
                   vector<Inhibiteur> const& vINS,
                   vector<Inhibiteur> const& vIS)
 : Reaction(infoReaction)
{   
    _inhibiteurs_NS = vINS;
    _inhibiteurs_S = vIS;
}

Reaction::Reaction(string const& ec, string const& nomEnzyme,
                   string const& organisme,
                   vector<string> const& substrats, 
                   vector<string> const& produits,
                   vector<float> const& km, vector<float> const& kcat,
                   string const& reversible,
                   vector<Inhibiteur> const& inhibiteurs)
{
    _ec = ec;
    _nomEnzyme = nomEnzyme;
    _organisme = organisme;
    _km = km;
    _kcat = kcat;
    _reversible = reversible == "r";
    
    
    // Stockage des substrats ET leurs stochiometrie !
    StockageMetabolite(substrats, "substrats");
        
    // Stockage des produits ET leurs stochiometrie
    StockageMetabolite(produits, "produits");

    // Stokage des inhibiteurs Spécifique (même organisme) ou non spécifique
    StockageInhibiteur(inhibiteurs);
}



/*
    Permet de remplir un vecteur de constantes avec les valeurs voulues
    constantes le vecteur à remplir
    
    constantes  constantes à remplir
    valeurs     les valeurs désirés des constantes
    delimiteur  sépare la valeur minimale de la maximale
*/
void Reaction::remplissageConstante(vector<float> & constantes, string valeurs, char const delimiteur)
{
    vector<string> split_str = split(valeurs, delimiteur);
    for(string valeur : split_str)
    {
        if(valeur != "?")
        {
            constantes.push_back(stof(valeur));
        }
        else // valeur == "?"
        {
            constantes.push_back(CONSTANTE_INCONNUE);
        }
    }
}


/*
    permet de remplir la la valeur de ph de la reaction
*/
void Reaction::remplissage_pH(string valeurs, char const delimiteur)
{
    vector<string> split_str = split(valeurs, delimiteur);
    if(split_str.size() == 2)
    {
        float valeur = (stof(split_str[0]) + stof(split_str[1]))/2;
        _ph.push_back(valeur);
    }
    else if(split_str.size() == 1)
    {
        if(split_str[0] == "?")
            _ph.push_back(CONSTANTE_INCONNUE);
        else
            _ph.push_back(stof(split_str[0]));
    }
    else
    {
        cerr << "Problem with one of the pH values in the file containing enzymatic reactions" << endl;
        exit(0);
    }
}


/*
    Stockage des substrats ou produits d'une reaction
    metabolites     vecteurs contenant les métabolites
    type            type des métabolites (substrats ou produits)
*/ 
void Reaction::StockageMetabolite(vector<string> const& metabolites, string const& type)
{
    vector<string> *conteneurMetabolites(0);
    vector<int> *stochiometrie(0);
    
    // Definition des variables de stockage
    if(type == "substrats")
    {
        conteneurMetabolites = &(_substrats);
        stochiometrie = &(_stochiometrieSubstrats);
    }
    else // type == "produits"
    {
        conteneurMetabolites = &(_produits);
        stochiometrie = &(_stochiometrieProduits);
    }
    
    // Stockage de l'information substrats ou produits
    for(size_t i = 0; i < metabolites.size(); i++)
    {
        size_t sepStochiometrie = metabolites[i].find('|');
        if(sepStochiometrie != string::npos)
        {
            string meta = metabolites[i].substr(sepStochiometrie+1, 
                                                string::npos);
            int stoc = stoi(metabolites[i].substr(0,sepStochiometrie));
            
            // Il ne faut pas prendre en compte la molecule H2O car forcement
            // en exces et intervient dans un TREEEEEEES grand nombre de 
            // réactions
            if( meta != "H20")
            {
                conteneurMetabolites->push_back(meta);
                stochiometrie->push_back(stoc);
            }
        }
        else // si pas de stochiometrie par defaut la valeur 1 est ajoutée
        {
            conteneurMetabolites->push_back(metabolites[i]);
            stochiometrie->push_back(1);
        }   
    }
}

/*
    Stockage des inhibiteurs connus d'une reaction
    inhibiteurs     vecteurs contenant les inhibiteurs
*/ 
void Reaction::StockageInhibiteur(vector<Inhibiteur> const& inhibiteurs)
{
    for(size_t i = 0; i < inhibiteurs.size(); i++)
    {
        if(inhibiteurs[i]._ec == _ec && 
           !inhibiteurs[i].inhibiteurConnu(_inhibiteurs_S))
        {
            if( inhibiteurs[i]._organisme == _organisme )
            {
                _inhibiteurs_S.push_back(inhibiteurs[i]);
            }
            else if(!inhibiteurs[i].inhibiteurConnu(_inhibiteurs_NS))
            {
                _inhibiteurs_NS.push_back(inhibiteurs[i]);
            }
        }
    }   
}



void Reaction::ajoutInhibitions_S(Inhibiteur inhibiteur)
{
    for(Inhibiteur const& i : _inhibiteurs_S)
    {
        if(i._nom == inhibiteur._nom)
            return;
    }
    _inhibiteurs_S.push_back(inhibiteur);
}
void Reaction::ajoutInhibitions_NS(Inhibiteur inhibiteur)
{
    for(Inhibiteur const& i : _inhibiteurs_NS)
    {
        if(i._nom == inhibiteur._nom)
            return;
    }
    _inhibiteurs_NS.push_back(inhibiteur);
}

/*
    vérifie que deux réactions sont prennent les mêmes substrats et 
    produisent les mêmes produits

    autre   l'autre réaction pour faire la comparaison

    retourne true si les deux réactions ont les mêmes substrats/produits
*/
bool Reaction::memeReaction(Reaction const& autre)
{
    if( memeContenu(_substrats, autre._substrats))
    {
        if( memeContenu(_produits, autre._produits))
        {
            return true;
        }
        else
        {
            return false;
        }
    }
    else
        return false;
}
bool Reaction::memeReaction(Reaction const* autre)
{
    if( memeContenu(_substrats, autre->_substrats))
    {
        if( memeContenu(_produits, autre->_produits))
        {
            return true;
        }
        else
        {
            return false;
        }
    }
    else
        return false;
}


bool Reaction::memeInhibiteurs(Reaction* autre)
{   
    if(_inhibiteurs_NS.size() == autre->_inhibiteurs_NS.size() && 
       _inhibiteurs_S.size() == autre->_inhibiteurs_S.size())
    {
        sort(_inhibiteurs_NS.begin(), _inhibiteurs_NS.end());
        sort(_inhibiteurs_S.begin(), _inhibiteurs_S.end());
        sort(autre->_inhibiteurs_NS.begin(), autre->_inhibiteurs_NS.end());
        sort(autre->_inhibiteurs_S.begin(), autre->_inhibiteurs_S.end());

        for( size_t i = 0; i < _inhibiteurs_NS.size(); i++)
        {
            if(_inhibiteurs_NS[i]._nom != autre->_inhibiteurs_NS[i]._nom)
                return false;
        }
        for( size_t i = 0; i < _inhibiteurs_S.size(); i++)
        {
            if(_inhibiteurs_S[i]._nom != autre->_inhibiteurs_S[i]._nom)
                return false;
        }
        return true;
    }
    else
        return false;
}

/*
    Vérifie que deux réactions sont différentes
    Si deux réactions ont le même code EC, le même organisme et les mêmes
    substrats/produits alors elles sont identiques. Même si les autres 
    éléments différents. Un certain nombre de réactions sont décritent dans
    plus d'une publication ce qui donne de la redondance d'information qui est
    gérée en ne prenant en compte qu'une seule d'entre elle.

    autre l'autre réaction pour faire la comparaison
*/
bool Reaction::operator!=(Reaction const& autre)
{
    if(_ec == autre._ec)
    {
        if(_organisme == autre._organisme)
        {
            if( memeContenu(_substrats, autre._substrats))
            {
                if( memeContenu(_produits, autre._produits))
                {
                    return false;
                }
                else
                    return true;
            }
            else
                return true;
        }
        else
            return true;
    }
    else
        return true;
}


/*
    Définition de l'opérateur < pour les réactions
    Seuls le code EC, l'organisme, et les substrats et produits sont pris en
    compte pour cette opérateur

    autre   l'autre réaction pour faire la comparaison

    retourn true si la reaction est < à l'autre
 */
bool Reaction::operator<(Reaction const& autre)
{
    if(_ec > autre._ec)
        return false;
    else if(_ec == autre._ec)
    {
        if(_organisme > autre._organisme)
            return false;
        else if(_organisme == autre._organisme)
        {
            // on regarde dans un premier temps le nombre de substrats/produits
            if(_substrats.size() > autre._substrats.size())
                return false;
            else if(_substrats.size() == autre._substrats.size())
            {
                if(_produits.size() > autre._produits.size())
                    return false;
                else if(_produits.size() == autre._produits.size())
                {
                    // Puis on s'interesse au contenu
                    if( memeContenu(_substrats, autre._substrats) &&
                        memeContenu(_produits, autre._produits))
                    {
                        return true;
                    }
                    else
                        return false;
                }
                else
                    return true;
            }
            else
                return true;
        }
        else
            return true;
    }
    else
        return true;
}