#include <iostream>
#include <algorithm> // find

#include "Recherche.hpp"
#include "Metabolite.hpp"
#include "fonctions.hpp"

using std::cerr;
using std::endl;
using std::ios;

extern int ci_string_compare (const string& s1, const string& s2);

Recherche::Recherche(Requete &requete)
{
    _requete = &requete;
    
    recherche();
}

/*
    Recherche des réactions pouvant connecter les entrées aux sorties
*/
void Recherche::recherche()
{   
    int niveau = 0;
    
    // s'il faut chercher des sorties
    if(_requete->_rechercheSortie)
    {
        // niveau 0 => profondeur 1 et/ou 2
        // niveau 1 => profondeur 3 et/ou 4
        while(niveau < _requete->_profondeur/2.0)
        {
            // Rechecher les inhibitions faites par les entrées
            if(_requete->_rechercheInhibitions)
                rechercheInhibitions(niveau);

            rechercheReactions(niveau);

            if(_requete->_rechercheInhibitions)
                rechercheInhibiteursSorties(niveau);
            niveau++;
        }   
    }
    // sinon si la recherche ne concerne que des entrées
    else
    {
        // niveau 0 => profondeur 1-2
        // niveau 1 => prodondeur 3-4
        while(niveau < _requete->_profondeur)
        {
            if(_requete->_rechercheInhibitions)
                rechercheInhibitions(niveau);
            rechercheReactions(niveau);
            niveau++;
        }
    }
}


//
// ========== RECHERCHE D'INHIBITEURS ==========
//

/*
    recherche les inhibitions qui doivent être prises en comptes
    
    niveau  la distance en nombre de réactions des entrées et des sortie 
*/
void Recherche::rechercheInhibitions(int niveau)
{
    // Variable qui stocke les potientiels inhibiteurs
    set<string> inhibiteursEntreesPotentiels;
    
    // remplissage de ces listes
    recherchePotentielsInhibiteurs(niveau, inhibiteursEntreesPotentiels);

    // Recherche des inhibitions induitent par ces potentiels inhibiteurs
    rechercheInhibitions_2(niveau, inhibiteursEntreesPotentiels);
    
}


/*
    Permet de d'établir une liste non redondante des potentielles
    inhibiteurs
    
    niveau      distance en nombre de réactions des entrées et des sorties

    inhibiteursEntreesPotentiels    liste des inhibiteurs d'entrées potentiels
*/
void Recherche::recherchePotentielsInhibiteurs(int niveau, 
                    set<string> & inhibiteursEntreesPotentiels)
{   
    if(niveau >= 1) // ajout des produits des réactions déja séléctionnées
    {
        // les produits des réactions d'entrées peuvent inhiber d'autres réactions
        // et donc les connecter
        for(Reaction const& reacEntrees : _reactionsEntrees[niveau-1])
        {
            for(string const& metabolite : reacEntrees._produits)
            {
                inhibiteursEntreesPotentiels.insert(metabolite);
            }
        }
    }
    else // niveau == 0, ajout uniquement des entrées
    {
        for(string const& metabolite : _requete->_entrees)
        {
            inhibiteursEntreesPotentiels.insert(metabolite);
        }
    }
}

/*
    Recherche dans la banques d'inhibitions, celles induitent par les inhibiteurs
    potentiels

    inhibiteursEntreesPotentiels    liste des inhibiteurs d'entrées potentiels
    inhibiteursSortiesPotentiels    liste des inhibiteurs de sorties potentiels
*/
void Recherche::rechercheInhibitions_2(int niveau, set<string> const& inhibiteursEntreesPotentiels)
{
    string ligne;
    vector<Inhibiteur> vecteurInhibitionsEntrees;
    vector<Inhibiteur> vecteurInhibitionsSorties;
    while(getline(_requete->_banqueInhibitions, ligne))
    {
        // vecteur qui contient les informations de l'inhibitions
        vector<string> inhibition = split(ligne, '\t');
        

        // récupérer les inhibiteurs 
        if(inhibiteursEntreesPotentiels.find(inhibition[2]) != 
           inhibiteursEntreesPotentiels.end())
        {
            Inhibiteur inhibiteur(inhibition);
            
            if(inhibiteurValide(inhibiteur))
            {
                vecteurInhibitionsEntrees.push_back(inhibiteur);

                if(niveau >= 1)
                {
                    for(Reaction & reaction : _reactionsEntrees[niveau-1])
                    {
                        for(string const& produit : reaction._produits)
                        {
							if (0 == ci_string_compare (produit, inhibiteur._nom))
                            {
                                Metabolite m(inhibiteur._nom, &reaction);
                                ajoutMetabolite(_reactionsEntreesAmonts, m);
                            }
                        }
                    }
                }
            }
        }
    }
    // Repositionnement du flag de banqueInhibitions au debut
    _requete->_banqueInhibitions.clear();
    _requete->_banqueInhibitions.seekg(0, ios::beg);
    
    // Suavegarde de la liste d'inhibiteur
    _inhibitionsEntrees.push_back(vecteurInhibitionsEntrees);
}


void Recherche::rechercheInhibiteursSorties(int niveau)
{
    // identifier les enzymes à inhiber
    set<string> enzymesSortieAInhiber;
    rechercheEnzymesSortieAInhiber(niveau, enzymesSortieAInhiber);

    // identifier les inhibiteurs de ces enzymes
    rechercheInhibiteursSorties_2(niveau, enzymesSortieAInhiber);
}

void Recherche::rechercheEnzymesSortieAInhiber(int niveau, set<string>& enzymesSortieAInhiber)
{
    for(Reaction const& reacSorties : _reactionsSorties[niveau])
    {
        enzymesSortieAInhiber.insert(reacSorties._ec);
    }
}

void Recherche::rechercheInhibiteursSorties_2(int niveau, set<string> const& enzymesSortieAInhiber)
{
    string ligne;
    vector<Inhibiteur> vecteurInhibitionsSorties;
    while(getline(_requete->_banqueInhibitions, ligne))
    {
        // vecteur qui contient les informations de l'inhibitions
        vector<string> inhibition = split(ligne, '\t');

        // récuperer les inhibiteur des réactions de sorties
        if(enzymesSortieAInhiber.find(inhibition[0]) != 
           enzymesSortieAInhiber.end())
        {
            Inhibiteur inhibiteur(inhibition);
            
            if(inhibiteurValide(inhibiteur))
            {                
                for(Reaction & reaction : _reactionsSorties[niveau])
                {
                    if(reaction._ec == inhibiteur._ec)
                    {
                        if(reaction._organisme == inhibiteur._organisme)
                        {
                            reaction.ajoutInhibitions_S(inhibiteur);
                        }
                        else
                        {
                            reaction.ajoutInhibitions_NS(inhibiteur);
                        }
                        Metabolite m(inhibiteur._nom, &reaction);
                        ajoutMetabolite(_reactionsSortiesAmonts, m);
                    }
                }

                vecteurInhibitionsSorties.push_back(inhibiteur);
            }
        }
    }
    // Repositionnement du flag de banqueInhibitions au debut
    _requete->_banqueInhibitions.clear();
    _requete->_banqueInhibitions.seekg(0, ios::beg);
    
    // Sauvegarde de la liste d'inhibiteur
    _inhibitionsSorties.push_back(vecteurInhibitionsSorties);
}

//
// ========== RECHERCHE DE REACTIONS ==========
//

/*
    Recherche des réactions d'interets
    
    niveau  nombre de réactions qui nous sépare des entrées et des sorties
*/
void Recherche::rechercheReactions(int niveau)
{
    // Pour stocker les réactions liées aux entrées ou sorties
    vector<Reaction> vecteurReactionsEntrees;
    vector<Reaction> vecteurReactionsSorties;
    
    // Recherche des substrats et produits cibles
    set<string> substratsPotentiels;
    set<string> produitsPotentiels;
    
    rechercheSubstratsProduits(niveau, substratsPotentiels, produitsPotentiels);
    
    // Lecture de la banque de réaction    
    string ligne; // une ligne correspond à une réaction
    while(getline(_requete->_banqueReactions, ligne))
    {    
        vector<string> infoReaction = split(ligne, '\t');
        
        // Si la réaction inhibée par un ou plusieurs inhibiteurs la stocker
        if(_requete->_rechercheInhibitions)
        {
            rechercheReactionsInhibees(niveau, infoReaction, 
                                       vecteurReactionsEntrees);

            rechercheReactionsProduisantInhibiteur(niveau, infoReaction,
                                                   vecteurReactionsSorties);
        }
        
        

        // Recherche des reactions consommant les substrats cibles
        rechercheReactionEntrees(infoReaction, substratsPotentiels,
                                 vecteurReactionsEntrees);
                                       
        // Recherche des reactions produitsant les produits cibles
        rechercheReactionSorties(infoReaction, produitsPotentiels,
                                 vecteurReactionsSorties);
            
    }
    // Stochage des reactions liées aux entrées ou sorties du niveau donné
    _reactionsEntrees.push_back(vecteurReactionsEntrees);
    _reactionsSorties.push_back(vecteurReactionsSorties);
    
    // Recommence la lecture de banque reaction au début
    _requete->_banqueReactions.clear();
    _requete->_banqueReactions.seekg(0, ios::beg);
}

/*
    Recherche des produits et substrats des réactions de niveau "niveau"
    
    niveau  nombre de réactions qui nous sépare des entrées et des sorties
    substratsPotentiels     liste des substrats potentiels à remplir
    produitsPotentiels      liste des produits potentiels à remplir
*/
void Recherche::rechercheSubstratsProduits( int niveau,
        set<string> &substratsPotentiels,
        set<string> &produitsPotentiels)
{
    if(niveau == 0)
    {
        for(string sub : _requete->_entrees)
        {
            substratsPotentiels.insert(sub);
        }
        for(string prod : _requete->_sorties)
        {
            produitsPotentiels.insert(prod);
        }
    }
    else // niveau > 0
    {
        for(Reaction& reac : _reactionsEntrees[niveau - 1])
        {
            for(string const& prod : reac._produits)
            {
                Metabolite m(prod, &reac);
                ajoutMetabolite(_reactionsEntreesAmonts, m);
                substratsPotentiels.insert(prod);
            }
        }   
        for(Reaction& reac : _reactionsSorties[niveau - 1])
        {
            for(string const& sub : reac._substrats)
            {
                Metabolite m(sub, &reac);
                ajoutMetabolite(_reactionsSortiesAmonts, m);
                produitsPotentiels.insert(sub);
            }
            for(Inhibiteur const& inhi : reac._inhibiteurs_S)
            {
                Metabolite m(inhi._nom, &reac);
                ajoutMetabolite(_reactionsSortiesAmonts, m);
                produitsPotentiels.insert(inhi._nom);
            }
            for(Inhibiteur const& inhi : reac._inhibiteurs_NS)
            {
                Metabolite m(inhi._nom, &reac);
                ajoutMetabolite(_reactionsSortiesAmonts, m);
                produitsPotentiels.insert(inhi._nom);
            }
        }
    }
}


/**
 *  \brief permet de stocker les réactions inhibées par les inhibiteurs cibles
 *
 *  \param niveau correspond a la distance en nombre de réactions des entrées
 *         ou des sorties
 *  \param infoReaction ligne de la banque de réactions splitée
 *  \param vecteurReactionsEntrees stocke les réactions entrees pour le niveau
 *         de recherche donné
 *  \param vecteurReactionsSorties stocke les réactions sorties pour le niveau
 *         de recherche donné
 */
void Recherche::rechercheReactionsInhibees(int niveau,
            vector<string> const& infoReaction,
            vector<Reaction> &vecteurReactionsEntrees)
{
    string const& code_ec   = infoReaction[0];
    string const& organisme = infoReaction[2];

    // vie = Vecteur Inhibiteurs d'Entrée
    vector<Inhibiteur> vie_specifique; 
    vector<Inhibiteur> vie_nonSpecifique;
    
    // stockage des réactions d'entrées inhibées
    estInhibee(code_ec, organisme,
               _inhibitionsEntrees[niveau], 
               vie_specifique, vie_nonSpecifique);
    
    // Si des inhibiteurs (specifique ou non) sont trouvés alors stocker la
    // réaction d'entrée
    if( vie_specifique.size() > 0 || vie_nonSpecifique.size() > 0)
    {
        Reaction reaction(infoReaction, vie_specifique, vie_nonSpecifique);

        if(reactionValide(reaction))
        {
            // pour charque inhibiteur stocker les reactions qui les produisent
            for(Inhibiteur const& i : vie_specifique)
            {

                for(Metabolite & m : _reactionsEntreesAmonts)
                {
					if (0 == ci_string_compare (m._nom, i._nom))
                    {
                        for(Reaction* ra : m._reactions_amonts)
                            reaction._reactions_amonts.push_back(ra);
                    }
                }
            }
            for(Inhibiteur const& i : vie_nonSpecifique)
            {

                for(Metabolite & m : _reactionsEntreesAmonts)
                {
					if (0 == ci_string_compare (m._nom, i._nom))
                    {
                        for(Reaction* ra : m._reactions_amonts)
                            reaction._reactions_amonts.push_back(ra);
                    }
                }
            }
            insertReaction(vecteurReactionsEntrees, reaction);
        }
    }
}

void Recherche::rechercheReactionsProduisantInhibiteur(int niveau,
                                                       vector<string> const& infoReaction,
                                                       vector<Reaction> & vecteurReactionsSorties)
{
    if(niveau >= 1)
    {
        vector<string> produits = split(infoReaction[4], "-+-");

        // recherche si au moins un des produits est un inhibiteur d'une reaction de niveau inférieur
        bool continuer = true;

        size_t t_produits = 0;
        size_t t_inhibiteur;
        while(t_produits < produits.size() && continuer)
        {
            string p = produits[t_produits];
            t_inhibiteur = 0;
            while(t_inhibiteur < _inhibitionsSorties[niveau-1].size() && continuer)
            {
                string i = _inhibitionsSorties[niveau-1][t_inhibiteur]._nom;
				if (0 == ci_string_compare (p, i))
                {
                    Reaction reaction(infoReaction);
                    // suite
                    if(reactionValide(reaction))
                    {
                        for(Metabolite & m : _reactionsSortiesAmonts)
                        {
							if (0 == ci_string_compare (m._nom, i))
                            {
                                for(Reaction* ra : m._reactions_amonts)
                                    reaction._reactions_amonts.push_back(ra);
                            }
                        }
                        insertReaction(vecteurReactionsSorties, reaction);
                        continuer = false;
                    }
                }
                ++t_inhibiteur;
            }
            ++t_produits;
        }
    }
}

/**
 *  \brief stockage des réactions repondant aux critères
 *
 *  \param infoReaction ligne de la banque de réactions splitée
 *  \param substratsPotentiels substrats potentiels des reactions d'entrees
 *  \param vecteurReactionsEntrees stocke les réactions d'entrées valides
 */
void Recherche::rechercheReactionEntrees(vector<string>   const& infoReaction,
                                         set<string>      const& substratsPotentiels,
                                         vector<Reaction> &      vecteurReactionsEntrees)
{
    // Creation de la reaction
    Reaction reaction(infoReaction);

    bool  insertion = false;
    
    for(string sub : reaction._substrats)
    {
        for(const string& sp : substratsPotentiels)
        {
            if(0 == ci_string_compare (sub, sp))
            {
                insertion = true;
                break;
            }
        }
    }
    if(insertion)
    {
        Reaction reaction(infoReaction);
        // Si la reaction est valide
        if(reactionValide(reaction))
        {
            // recherche de celles selectionné qui produisent les substrats
            set<Reaction*> reactions_amonts = rechercheAmont(
                                                 reaction._substrats,
                                                 _reactionsEntreesAmonts);
            
            
            for(Reaction* r : reactions_amonts)
            {
                reaction._reactions_amonts.push_back(r);
            }
            
            // insertion de la reaction
            insertReaction(vecteurReactionsEntrees, reaction);
        }
    }
}

/**
 *  \brief stockage des réactions repondant aux critères
 *
 *  \param infoReaction ligne de la banque de réactions splitée
 *  \param produitsPotentiels substrats potentiels des reactions de sorties
 *  \param vecteurReactionsSorties stocke les réactions de sorties valides
 */
void Recherche::rechercheReactionSorties(vector<string>   const& infoReaction,
                                         set<string>      const& produitsPotentiels,
                                         vector<Reaction> &      vecteurReactionsSorties)
{
    // Creation de la reaction
    Reaction reaction(infoReaction);
    bool  insertion = false;
    

    for(string prod : reaction._produits)
    {
        // Si un des produits de la reactions est un produitsPotentiels
        // stocker la réaction
        for(const string& pp : produitsPotentiels)
        {
            if(0 == ci_string_compare (prod, pp))
            {
                insertion = true;
                break;
            }
        }
    }
    if(insertion)
    {
        Reaction reaction(infoReaction);
        // Si la reaction est valide
        if( reactionValide(reaction))
        {
            // recherche de celles selectionné qui consomment les produits
            set<Reaction*> reactions_amonts = rechercheAmont(
                                                 reaction._produits,
                                                 _reactionsSortiesAmonts);
            
            
            for(Reaction* r : reactions_amonts)
            {
                reaction._reactions_amonts.push_back(r);
            }
            
            // insertion de la reaction
            insertReaction(vecteurReactionsSorties, reaction);
        }
    }
}


/**
 *  \brief Teste si la réaction correspond aux critéres de la requete
 *
 *  \param reaction la reaction à tester
 */

bool Recherche::reactionValide(Reaction & reaction)
{
    vector<string>::iterator itv;
    
    // test: si la reaction ne consomme pas de sorties
    for(string sub : reaction._substrats)
    {
        for(const string& sortie : _requete->_sorties)
        {
            if (0 == ci_string_compare(sortie, sub))
                return false;
        }
    }
    
    // test: si la reaction ne produit pas des entrées
    for(string prod : reaction._produits)
    {        
        for(const string& entree : _requete->_entrees)
        {
            if (0 == ci_string_compare (entree, prod))
                return false;
        }
    }
    
    // test: si la reaction a des substrats inconnus
    for(const string& sub : reaction._substrats)
    {
        if (0 == ci_string_compare (sub, "?"))
            return false;
    }
    
    // test: si la reaction a des produits inconnus
    for(const string& prod : reaction._produits)
    {
        if (0 == ci_string_compare(prod, "?"))
            return false;
    }
    
    // test: si les km sont dans l'intrvalle voulut
    if(_requete->_kmMin != CONSTANTE_MIN ||
       _requete->_kmMax != CONSTANTE_MAX)
    {
        for(float km : reaction._km)
        {
            if( km < _requete->_kmMin || km > _requete->_kmMax)
                return false;
        }
    }
    
    // test: si les kcat sont dans l'intrvalle voulut
    if(_requete->_kcatMin != CONSTANTE_MIN ||
       _requete->_kcatMax != CONSTANTE_MAX)
    {
        for(float kcat : reaction._kcat)
        {
            if( kcat < _requete->_kcatMin || kcat > _requete->_kcatMax)
                return false;
        }
    }
    
    // test: si les ph favorables sont dans l'intrvalle voulut
    if(_requete->_phMin != CONSTANTE_MIN ||
       _requete->_phMax != CONSTANTE_MAX)
    {
        for(float ph : reaction._ph)
        {
            if( ph < _requete->_phMin || ph > _requete->_phMax)
                return false;
        }
    }
    
    // si aucun problème retourner true
    return true;
}

/**
 *  \brief Teste si l'inhibiteur correspond aux critéres de la requete
 *
 *  \param inhibiteur à tester
 */
bool Recherche::inhibiteurValide(Inhibiteur & inhibiteur)
{  
    // test: si le ki est dans l'intervalle voulut
    if(_requete->_kiMin != CONSTANTE_MIN ||
       _requete->_kiMax != CONSTANTE_MAX)
    {
        if( inhibiteur._ki < _requete->_kiMin || inhibiteur._ki > _requete->_kiMax)
            return false;
    }
    
    // test: si l'IC50 est dans l'intervalle
    if(_requete->_ic50Min != CONSTANTE_MIN ||
       _requete->_ic50Max != CONSTANTE_MAX)
    {
        if( inhibiteur._ic50 < _requete->_ic50Min || inhibiteur._ic50 > _requete->_ic50Max)
            return false;
    }
    
    // si aucun problème retourner true
    return true;
}


set<Reaction*> Recherche::rechercheAmont( 
                                  vector<string> const& metabolitesCibles,
                                  vector<Metabolite> const& vMetabolite)
{
    set<Reaction*> vReaction;
    for(string m1 : metabolitesCibles)
    {
        for(Metabolite const& m2 : vMetabolite)
        {
			if (0 == ci_string_compare (m1, m2._nom))
            {
                for(Reaction* pReac : m2._reactions_amonts)
                {
                    vReaction.insert(pReac);
                }
            }
        }
    }
    return vReaction;
}



/*
    Affichage du nombre de réactions trouvées
*/
void Recherche::affichage()
{
    int nbRE = 0;
    int nbRS = 0;
    
    switch(_requete->_profondeur)
    {       
        case 4:
            nbRS += _reactionsSorties[1].size();
        case 3:     
            nbRE += _reactionsEntrees[1].size();
        case 2:     
            nbRS += _reactionsSorties[0].size();
        case 1:
            nbRE += _reactionsEntrees[0].size();
            
    }
    cerr << "Number of input reactions =" << nbRE << endl;
    cerr << "- level 0" << endl;
    int n = 1;
    for(vector<Reaction> const& vr : _reactionsEntrees)
    {
        cerr << "- profondeur " << n++ << endl;
        for(Reaction const& r : vr)
        {
            cerr << r._ec << "\t";
            for(string const& s : r._substrats)
                cerr << s << "|";
            cerr << "\t";
            for(string const& p : r._produits)
                cerr << p << "|";
            cerr << endl;
        }
    }
    cerr << "Number of output reactions =" << nbRS << endl;
    for(vector<Reaction> const& vr : _reactionsSorties)
    {
        cerr << "- distance " << n++ << endl;
        for(Reaction const& r : vr)
        {
            cerr << r._ec << "\t";
            for(string const& s : r._substrats)
                cerr << s << "|";
            cerr << "\t";
            for(string const& p : r._produits)
                cerr << p << "|";
            cerr << endl;
        }
    }
}